@@ -54,6 +54,11 @@ class BwCircuitStat:
     self.is_hs = is_hs
     self.is_service = 1
     self.is_hsdir = 0
+    self.dropped_cells_allowed = 0
+    self.purpose = None
+    self.hs_state = None
+    self.old_purpose = None
+    self.old_hs_state = None
     self.in_use = 0
     self.built = 0
     self.created_at = time.time()
@@ -234,6 +239,30 @@ class BandwidthStats:
         self.circs[event.id].is_hsdir = 1
       plog("DEBUG", "Added circ for "+event.raw_content())
 
+    # Debugging
+    self.circs[event.id].purpose = event.purpose
+    self.circs[event.id].hs_state = event.hs_state
+
+    # Workaround for Tor bug #29700 (see also
+    # https://github.com/mikeperry-tor/vanguards/issues/37).
+    # Class 3: Service rend circs can sometimes fail ntor handshake on extend
+    if event.purpose == "HS_SERVICE_REND":
+      if event.hs_state == "HSSR_CONNECTING":
+        self.circs[event.id].dropped_cells_allowed = 1
+      else:
+        self.circs[event.id].dropped_cells_allowed = 0
+    # Workaround for Tor bug #29927 (see also
+    # https://github.com/mikeperry-tor/vanguards/issues/37).
+    # Class 4: Mysterious client-side cases of dropped cells
+    # and protocol errors.
+    elif event.purpose == "HS_CLIENT_REND":
+      self.circs[event.id].dropped_cells_allowed = 1
+    elif event.purpose == "HS_CLIENT_INTRO":
+      if event.hs_state == "HSCI_DONE":
+        self.circs[event.id].dropped_cells_allowed = 1
+      else:
+        self.circs[event.id].dropped_cells_allowed = 0
+
     # Consider all BUILT circs that have a specific HS purpose
     # to be "in_use".
     if event.status == stem.CircStatus.BUILT or \
@@ -253,7 +282,6 @@ class BandwidthStats:
         self.circs[event.id].guard_fp = event.path[0][0]
         plog("DEBUG", "Circ "+event.id+" now in-use. %d delivered bytes.",
              self.circs[event.id].delivered_read_bytes)
-
     # Extending a circuit means the network is OK
     elif event.status == "EXTENDED":
       if self.disconnected_circs:
@@ -269,6 +297,11 @@ class BandwidthStats:
     if event.id not in self.circs:
       return
 
+    self.circs[event.id].purpose = event.purpose
+    self.circs[event.id].hs_state = event.hs_state
+    self.circs[event.id].old_purpose = event.old_purpose
+    self.circs[event.id].old_hs_state = event.old_hs_state
+
     if event.purpose[0:9] == "HS_CLIENT":
       self.circs[event.id].is_service = 0
     elif event.purpose[0:10] == "HS_SERVICE":
@@ -285,6 +318,44 @@ class BandwidthStats:
         plog("DEBUG", "Circ "+event.id+" now in-use. %d delivered bytes.",
              self.circs[event.id].delivered_read_bytes)
 
+      # Tor Bug #29786 workaround (see also
+      # https://github.com/mikeperry-tor/vanguards/issues/37)
+      if event.purpose == "PATH_BIAS_TESTING":
+        # If purpose changes to PATH_BIAS_TESTING, then check for known cases
+        # of dropped cells:
+        # 1a. Path bias: pending RELAY_COMMAND_INTRO_ESTABLISHED cell
+        if event.old_purpose == "HS_SERVICE_INTRO" \
+           and event.old_hs_state == "HSSI_CONNECTING":
+          self.circs[event.id].dropped_cells_allowed = 1
+        # 1b. Path bias: pending RELAY_COMMAND_RENDEZVOUS_ESTABLISHED cell
+        if event.old_purpose == "HS_CLIENT_REND" \
+           and event.old_hs_state == "HSCR_CONNECTING":
+          self.circs[event.id].dropped_cells_allowed = 1
+        # 1c. Path bias: pending RELAY_COMMAND_INTRODUCE_ACK cell
+        if event.old_purpose == "HS_CLIENT_INTRO" \
+           and event.old_hs_state != "HSCI_DONE":
+          self.circs[event.id].dropped_cells_allowed = 1
+      # Workaround for Tor bug #29700 (see also
+      # https://github.com/mikeperry-tor/vanguards/issues/37).
+      # Class 3: Service rend circs can sometimes fail ntor handshake on extend
+      elif event.purpose == "HS_SERVICE_REND":
+        if event.hs_state == "HSSR_CONNECTING":
+          self.circs[event.id].dropped_cells_allowed = 1
+        else:
+          self.circs[event.id].dropped_cells_allowed = 0
+      # Workaround for Tor bug #29927 (see also
+      # https://github.com/mikeperry-tor/vanguards/issues/37).
+      # Class 4: Mysterious client-side cases of dropped cells
+      # and protocol errors.
+      elif event.purpose == "HS_CLIENT_REND":
+        self.circs[event.id].dropped_cells_allowed = 1
+      elif event.purpose == "HS_CLIENT_INTRO":
+        if event.hs_state == "HSCI_DONE":
+          self.circs[event.id].dropped_cells_allowed = 1
+        else:
+          self.circs[event.id].dropped_cells_allowed = 0
+
+
     plog("DEBUG", event.raw_content())
 
   def circbw_event(self, event):
@@ -370,7 +441,7 @@ class BandwidthStats:
     for circ in kill_circs:
       self.limit_exceeded("NOTICE", "CIRC_MAX_AGE_HOURS",
                           circ.circ_id,
-                          now - circ.created_at,
+                          (now - circ.created_at)/_SECS_PER_HOUR,
                           CIRC_MAX_AGE_HOURS)
       control.try_close_circuit(self.controller, circ.circ_id)
 
@@ -381,11 +452,52 @@ class BandwidthStats:
     self.check_circ_ages(now)
 
   def check_circuit_limits(self, circ):
-    # Now that all supported tors have #25573, any dropped cell is bad.
-    if circ.dropped_read_cells() > 0:
-      plog("WARN",
-           "Possible attack! Got a dropped cell on circ %s.", circ.circ_id)
-      control.try_close_circuit(self.controller, circ.circ_id)
+    if circ.dropped_read_cells() > circ.dropped_cells_allowed:
+      # Workaround for Tor bug #29699 (see also
+      # https://github.com/mikeperry-tor/vanguards/issues/37).
+      # Case 2. Intro circs can get dupped cells on retries while in use.
+      # Demote log to notice, but still close them though, cause fuck it.
+      # They get rebuilt, and side channels are bad, mmmk.
+      if circ.purpose == "HS_SERVICE_INTRO" and \
+         circ.hs_state == "HSSI_ESTABLISHED":
+        plog("NOTICE", "Tor bug #29699: Got %d dropped cell on circ %s "\
+                       +"(in state %s %s; old state %s %s).",
+                       circ.dropped_read_cells(), circ.circ_id,
+                       str(circ.purpose), str(circ.hs_state),
+                       str(circ.old_purpose), str(circ.old_hs_state))
+        control.try_close_circuit(self.controller, circ.circ_id)
+      elif not circ.built:
+       plog("INFO", "Tor bug #29927: Got a dropped cell on circ %s "\
+                       +"(in state %s %s; old state %s %s).", circ.circ_id,
+                       str(circ.purpose), str(circ.hs_state),
+                       str(circ.old_purpose), str(circ.old_hs_state))
+      else:
+        plog("WARN", "Possible Tor bug, or possible attack if very frequent: "\
+                     +"Got %d dropped cell on circ %s "\
+                     +"(in state %s %s; old state %s %s)",
+                     circ.dropped_read_cells(), circ.circ_id,
+                     str(circ.purpose), str(circ.hs_state),
+                     str(circ.old_purpose), str(circ.old_hs_state))
+        control.try_close_circuit(self.controller, circ.circ_id)
+    elif circ.dropped_read_cells() > 0:
+      # Log workaround drop cell cases for completeness
+      if circ.purpose == "PATH_BIAS_TESTING":
+        plog("INFO", "Tor bug #29786: Got a dropped cell on circ %s "\
+                       +"(in state %s %s; old state %s %s).", circ.circ_id,
+                       str(circ.purpose), str(circ.hs_state),
+                       str(circ.old_purpose), str(circ.old_hs_state))
+      elif circ.purpose == "HS_SERVICE_REND" and \
+           circ.hs_state == "HSSR_CONNECTING":
+        plog("INFO", "Tor bug #29700: Got a dropped cell on circ %s "\
+                       +"(in state %s %s; old state %s %s).", circ.circ_id,
+                       str(circ.purpose), str(circ.hs_state),
+                       str(circ.old_purpose), str(circ.old_hs_state))
+      else:
+        plog("INFO", "Tor bug #29927: Got a dropped cell on circ %s "\
+                       +"(in state %s %s; old state %s %s).", circ.circ_id,
+                       str(circ.purpose), str(circ.hs_state),
+                       str(circ.old_purpose), str(circ.old_hs_state))
+
     if CIRC_MAX_MEGABYTES > 0 and \
        circ.total_bytes() > CIRC_MAX_MEGABYTES*_BYTES_PER_MB:
       self.limit_exceeded("NOTICE", "CIRC_MAX_MEGABYTES",
