@@ -4,6 +4,10 @@ import getpass
 
 from .logger import plog
 
+from . import __version__
+
+_CLOSE_CIRCUITS = True
+
 def authenticate_any(controller, passwd=""):
   try:
     controller.authenticate()
@@ -20,7 +24,8 @@ def authenticate_any(controller, passwd=""):
     print("Unable to authenticate: %s" % exc)
     sys.exit(1)
 
-  plog("NOTICE", "Connected to Tor version %s" % controller.get_version())
+  plog("NOTICE", "Vanguards %s connected to Tor %s using stem %s",
+       __version__, controller.get_version(), stem.__version__)
 
 def get_consensus_weights(consensus_filename):
   parsed_consensus = next(stem.descriptor.parse_file(consensus_filename,
@@ -31,9 +36,9 @@ def get_consensus_weights(consensus_filename):
   return parsed_consensus.bandwidth_weights
 
 def try_close_circuit(controller, circ_id):
-  try:
-    controller.close_circuit(circ_id)
-    plog("NOTICE", "We force-closed circuit "+str(circ_id))
-  except stem.InvalidRequest as e:
-    plog("INFO", "Failed to close circuit "+str(circ_id)+": "+str(e.message))
-
+  if _CLOSE_CIRCUITS:
+    try:
+      controller.close_circuit(circ_id)
+      plog("NOTICE", "We force-closed circuit "+str(circ_id))
+    except stem.InvalidRequest as e:
+      plog("INFO", "Failed to close circuit "+str(circ_id)+": "+str(e.message))
